from typing import Tuple, List, Union
import torch
from ._lib import LLMatcher, LLExecutor, LLInterpreter


def get_bitmask_shape(batch_size: int, vocab_size: int) -> Tuple[int, int]:
    return (batch_size, (vocab_size + 31) // 32)


def allocate_token_bitmask(batch_size: int, vocab_size: int) -> torch.Tensor:
    return torch.full(
        get_bitmask_shape(batch_size, vocab_size),
        -1,
        dtype=torch.int32,
        pin_memory=torch.cuda.is_available(),
    )


@torch.compile(dynamic=True)  # faster than dynamic=False and jit.script
def apply_token_bitmask_inplace_kernel(logits: torch.Tensor,
                                       mask: torch.Tensor) -> None:
    mask_expanded = torch.repeat_interleave(mask, 32, dim=1)
    bit_indices = torch.arange(32, device=logits.device,
                               dtype=torch.int32).repeat(mask.shape[1])
    bit_masks = (mask_expanded >> bit_indices) & 1  # Extract each bit
    bit_masks = bit_masks[:, :logits.shape[1]]  # Trim to match vocab size
    logits.masked_fill_(bit_masks == 0, float("-inf"))  # Apply mask


def apply_token_bitmask_inplace(logits: torch.Tensor,
                                mask: torch.Tensor) -> None:
    if logits.dim() == 1:
        logits = logits.unsqueeze(0)
    if mask.dim() == 1:
        mask = mask.unsqueeze(0)
    assert mask.dtype == torch.int32, "Mask must be int32"
    assert logits.dim() == 2, "Logits must be 2D"
    batch, vocab = logits.shape
    m_batch, m_vocab = mask.shape
    assert batch == m_batch, "Batch size mismatch"
    cutoff = 32 * m_vocab
    if vocab > cutoff:
        logits[:, cutoff:] = float("-inf")
        logits = logits[:, :cutoff]
    apply_token_bitmask_inplace_kernel(logits, mask)


def fill_next_token_bitmask(interp: Union[LLMatcher, LLInterpreter],
                            bitmask: torch.Tensor,
                            index: int = 0) -> None:
    assert bitmask.dtype == torch.int32, "Mask must be int32"
    assert bitmask.is_cpu, "Mask must be on CPU"
    assert bitmask.dim() == 2, "Mask must be 2D"
    v = bitmask[index, :]
    assert v.is_contiguous(), "Mask must be contiguous"
    interp.unsafe_compute_mask_ptr(v.data_ptr(), v.numel() * v.element_size())


def fill_next_token_bitmask_par(executor: LLExecutor,
                                matchers: List[Tuple[LLMatcher, int]],
                                bitmask: torch.Tensor) -> None:
    assert bitmask.dtype == torch.int32, "Mask must be int32"
    assert bitmask.is_cpu, "Mask must be on CPU"
    assert bitmask.dim() == 2, "Mask must be 2D"
    batch, vocab = bitmask.shape
    assert bitmask.is_contiguous(), "Mask must be contiguous"
    executor.unsafe_compute_mask_ptr(matchers, bitmask.data_ptr(), vocab * 4,
                                     batch)



import llguidance.llamacpp
import llama_cpp
import os
import requests # type: ignore
from typing import Any

def get_llama_vocab_file(pytestconfig: Any) -> str:
    url = "https://raw.githubusercontent.com/ggml-org/llama.cpp/f4ab2a41476600a98067a9474ea8f9e6db41bcfa/models/ggml-vocab-llama-bpe.gguf"
    cache_dir = pytestconfig.cache.makedir("llama_vocab")
    file_name = "vocab.gguf"
    file_path = os.path.join(cache_dir, file_name)

    if not os.path.exists(file_path):
        r = requests.get(url)
        r.raise_for_status()
        with open(file_path, "wb") as f:
            f.write(r.content)

    return file_path


def test_llama_cpp(pytestconfig: Any) -> None:
    filepath = get_llama_vocab_file(pytestconfig)
    p = llama_cpp.llama_model_default_params()
    p.vocab_only = True
    model = llama_cpp.llama_model_load_from_file(filepath.encode(), p)
    assert model is not None
    vocab = llama_cpp.llama_model_get_vocab(model)
    assert vocab is not None
    llt = llguidance.llamacpp.lltokenizer_from_vocab(vocab)
    for s in [
            "Hello world!", "Hello world! こんにちは世界！", "wave 👋", "heart 👋💖",
            "1`a`b`c`d`e`f`g`h`i"
    ]:
        toks = llt.tokenize_str(s)
        print(llt.dbg_tokens(toks))
        assert llt.decode_str(toks) == s
    toks = llt.tokenize_bytes(b"\x8b")
    print(llt.dbg_tokens(toks))
    print(toks)
    assert len(toks) == 1
    assert llt.decode_bytes(toks) == b"\x8b"

    toks1 = llt.tokenize_str("<|eot_id|>")
    toks0 = llt.tokenize_str("<|eot_id|>", parse_special=False)
    assert toks1 == toks0
    assert len(toks0) > 1
    toks2 = llt.tokenize_str("<|eot_id|>", parse_special=True)
    assert len(toks2) == 1

    toks3 = llt.tokenize_str("a<|eot_id|>b", parse_special=True)
    print(llt.dbg_tokens(toks3))
    assert len(toks3) == 3



import llguidance as llg
from transformers import AutoTokenizer
from huggingface_hub import hf_hub_download

model_name = "unsloth/Meta-Llama-3.1-8B-Instruct"
tokenizer = AutoTokenizer.from_pretrained(model_name)

tokenizer_json_path = hf_hub_download(
    repo_id=model_name, filename="tokenizer.json"
)
with open(tokenizer_json_path, "r") as f:
    llg_tokenizer = llg.LLTokenizer(f.read())

def tokenize(s: str) -> list[int]:
    r = tokenizer(s, return_tensors="pt", add_special_tokens=False)
    return r["input_ids"][0].tolist()


def test_tokenize(s: str):
    t = tokenize(s)
    dbg = llg_tokenizer.dbg_tokens(t)
    print(f"{repr(s)} -> {t} {dbg}")


test_tokenize('additional_properties')
test_tokenize('{"_properties')
test_tokenize('<|end_of_text|>_properties')
test_tokenize('_properties')
test_tokenize('\u0002_properties')



#!/usr/bin/env python3

import subprocess
import re
import sys


def get_version_from_cargo_toml(cargo_toml_path):
    try:
        with open(cargo_toml_path, "r") as file:
            cargo_toml_contents = file.read()

        # Match version field in Cargo.toml
        match = re.search(r'\bversion\s*=\s*"([^"]+)"', cargo_toml_contents)

        if match:
            return match.group(1)
        else:
            print(f"Error: Version not found in {cargo_toml_path}.")
            return None
    except FileNotFoundError:
        print(f"Error: {cargo_toml_path} not found.")
        return None
    except Exception as e:
        print(f"Error reading {cargo_toml_path}: {e}")
        return None


def update_cargo_toml(cargo_toml_path, derivre_version):
    try:
        with open(cargo_toml_path, "r") as file:
            cargo_toml_contents = file.read()

        # Patterns for replacing the version in Cargo.toml
        derivre_pattern = r'(derivre\s*=\s*\{[^}]*version\s*=\s*")[^"]*(")'

        cargo_toml_contents = re.sub(
            derivre_pattern,
            lambda m: m.group(1) + derivre_version + m.group(2),
            cargo_toml_contents,
        )

        # Write the updated contents back to the Cargo.toml file
        with open(cargo_toml_path, "w") as file:
            file.write(cargo_toml_contents)

        print(f"{cargo_toml_path} updated successfully.")
    except FileNotFoundError:
        print(f"Error: {cargo_toml_path} not found.")
    except Exception as e:
        print(f"Error updating {cargo_toml_path}: {e}")


# Get the version from ../derivre/Cargo.toml
derivre_version = get_version_from_cargo_toml("../derivre/Cargo.toml")

# Check if the version was retrieved successfully
if not derivre_version:
    print("Error retrieving version. Exiting.")
    sys.exit(1)

# List of Cargo.toml paths to update
cargo_toml_paths = [
    "parser/Cargo.toml",
]

# Update each Cargo.toml file
for cargo_toml_path in cargo_toml_paths:
    update_cargo_toml(cargo_toml_path, "=" + derivre_version)

subprocess.run(["cargo", "check"], check=True)

print("All Cargo.toml files updated and cargo fetch run successfully.")



from typing import List, Optional, TYPE_CHECKING

from ._lib import LLTokenizer

if TYPE_CHECKING:
    import tiktoken


def lltokenizer_from_encoding(
    encoding: 'tiktoken.Encoding',
    *,
    n_vocab: Optional[int] = None,
    eos_token: Optional[int] = None,
    slices: Optional[List[str]] = None,
) -> LLTokenizer:
    """
    Create a new tokenizer from a tiktoken Encoding object.
    This is an expensive operation (~1s), so the result should be cached.

    Args:
        encoding: tiktoken.Encoding - the encoding object to use
        n_vocab: int - override the size of the vocabulary
        eos_token: int - override the EOS token
        slices: List[str] - configuration for slicer optimization; pass [] to disable,
            or None to use the default configuration
    """

    return LLTokenizer.from_tiktoken(  # type: ignore
        encoder=encoding._mergeable_ranks,
        special_tokens=encoding._special_tokens,
        pattern=encoding._pat_str,
        eos_token=encoding.eot_token if eos_token is None else eos_token,
        n_vocab=n_vocab,
        slices=slices)



#!/usr/bin/env python3

import sys
import os
import re

exclude = {"tmp/", "node_modules/", "target/"}

# Read .gitmodules and extract paths to exclude
if os.path.exists('.gitmodules'):
    with open('.gitmodules', 'r', encoding='utf-8') as f:
        for match in re.finditer(r'^\s*path = (.*)$', f.read(), re.MULTILINE):
            exclude.add(match.group(1) + "/")

links = {}
files = {}

# Read files from arguments
for file in sys.argv[1:]:
    if any(file.startswith(ex) for ex in exclude):
        continue
    
    abs_file = os.path.abspath(file)
    with open(abs_file, 'r', encoding='utf-8') as f:
        content = f.read()
        files[abs_file] = content
    
    # Mask text inside triple backticks to preserve line numbers
    masked_content = re.sub(r'```.*?```', lambda m: '\n'.join(['' for _ in m.group(0).split('\n')]), content, flags=re.DOTALL)
    
    for match in re.finditer(r'^#+ (.*)$', masked_content, re.MULTILINE):
        title = match.group(1)
        anchor = "#" + re.sub(r'[^a-z0-9 \-]+', '', title.lower()).replace(' ', '-')
        links[abs_file + anchor] = True

numerr = 0
numlinks = 0
numanchors = 0
numhttp = 0

for filename, content in files.items():
    line_no = 0
    
    # Mask text inside triple backticks to preserve line numbers
    masked_content = re.sub(r'```.*?```', lambda m: '\n'.join(['' for _ in m.group(0).split('\n')]), content, flags=re.DOTALL)
    
    for line in masked_content.split("\n"):
        line_no += 1
        
        for match in re.finditer(r'\[([^\]]+)\]\(([^\)]+)\)', line):
            title, link = match.groups()
            
            if link.startswith(("https://", "http://", "mailto:")):
                numhttp += 1
                continue
            
            numlinks += 1
            if link.startswith("#"):
                link = filename + link
            
            linkfile, *anchor_parts = link.split("#", 1)
            linkfile = os.path.abspath(os.path.join(os.path.dirname(filename), linkfile))
            anchor = "#" + anchor_parts[0] if anchor_parts and not anchor_parts[0].startswith("L") else ""
            line_ref = anchor_parts[0] if anchor_parts and anchor_parts[0].startswith("L") else ""
            
            if not os.path.exists(linkfile):
                numerr += 1
                print(f"{filename}:{line_no}: Broken link '{title}': {link}")
                continue
            
            if anchor and linkfile + anchor not in links:
                numerr += 1
                print(f"{filename}:{line_no}: Broken link to anchor '{title}': {link}")
            elif line_ref:
                try:
                    line_number = int(line_ref[1:])
                    with open(linkfile, 'r', encoding='utf-8') as f:
                        lines = f.readlines()
                        if line_number > len(lines) or line_number <= 0:
                            numerr += 1
                            print(f"{filename}:{line_no}: Broken link to line '{title}': {link}")
                except ValueError:
                    numerr += 1
                    print(f"{filename}:{line_no}: Invalid line reference '{title}': {link}")

if numerr > 0:
    print(f"Found {numerr} broken links")
    sys.exit(1)
else:
    print(f"Exclude: {', '.join(exclude)}")
    print(f"Checked {numlinks} links (incl. {numanchors} anchors). Skipped {numhttp} http links.")